/* --- Copyright University of Birmingham 1995. All rights reserved. ------
 > File:            $poplocal/local/lib/advent_objectclass.p
 > Purpose:         Explained in TEACH * ADVENT_OBJECTCLASS
 > Author:          Original version (for LIB OBJ) Chris Mellish, Feb 1985 (see revisions)
 >  Updated for LIB NEWOBJ (with DEFAULTS) Aaron Sloman, June 1985
 > 		(See TEACH ADVENT.NEWOBJ)
 >
 >  Updated for LIB OBJECTCLASS Luc Beaudoin November 1992
 >  Further updated for latest Objectclass Aaron Sloman November 1993
 >
 > Documentation: TEACH * ADVENT_OBJECTCLASS
 > Related Files: LIB * OBJECTCLASS, TEACH * OBJECTCLASS_EXAMPLE
 */

/*
Example programming of a simulated "adventure world" using LIB OBJECTCLASS

To make this your own file, called 'advent.p' do:
            <ENTER> name advent.p
*/

uses objectclass;

/* ---------- DEFINING THE MAIN CLASSES ---------- */

;;; For ease of reading we use a mixture of upper and lower case in
;;;	class names and some method names (this is not necessary).

;;; Also to prevent confusion with other programs or packages we use
;;; the prefix "Ao" or "ao" to start off identifiers from the advent_objectlass
;;; library. "Ao" is used in class names, "ao" in slot or method names.


;;; THE TOP LEVEL CLASS

;;; Set up a super-class of all classes in this file. All other classes
;;; will inherit the slot "aoName" from this.

define :class AoObject;
    slot aoName;
enddefine;

;;; Adventures happen in houses. Define a class for houses. In the
;;; simplest game there is only one instance, but there could be
;;; several houses, in principle, which is why this is a class.

define :class AoHouse;
    is AoObject;
    slot aoPlayers  ==  [];
    slot aoStartloc;
    slot aoRooms;
enddefine;

;;; The following class is for treasures, etc. which players may pick
;;; up as they wander around

define :class AoPhysicalObject;
    is AoObject;
    slot aoValue  ==  0;
    slot aoWeight;
enddefine;

;;; The "adventure world" consists of a number of "room" objects
;;; in a house. Each room has four walls, on the North, South, East
;;; and West sides. Each wall has doors leading to adjoining rooms.
;;; There may be some objects in the room, as well as current Players.

define :class AoRoom;
    is AoObject;
    slot aoPlayers == [];        ;;; The players currently in this room
    slot aoNorth == false;       ;;; )
    slot aoSouth == false;       ;;; ) The adjoining rooms in various
    slot aoEast == false;        ;;; )   directions
    slot aoWest == false;        ;;; )
    slot aoObjects ==  [];       ;;; The objects lying in the room
    slot aoHouse;                ;;; a room knows its own house
enddefine;


;;; A SUPERCLASS FOR ALL ACTIVE AGENTS, I.E. PLAYERS.
;;; Players have a state of health, some current possessions, and
;;; a current location in the form of a room.

define :class AoPlayer;
    is AoObject;
    slot aoHealth  ==  100;      ;;; Lies between 0 and 100, approximately
    slot aoPossessions  ==  [];  ;;; Objects currently carried by the player
    slot aoRoom;                 ;;; Room the player is in
enddefine;

;;; Now produce a variety of different subclasses. They will behave
;;; differently when methods are applied to them.

define :class AoGreedy;
    is AoPlayer;
enddefine;

define :class AoVicious;
    is AoPlayer;
enddefine;


/* --- DEFINE SOME METHODS FOR MAKING THINGS HAPPEN IN THE GAME --- */

;;; Declare a procedure variable that will later be defined as a method
;;; identifier

vars procedure aoIntroduce;

;;; First a method to make a player enter a house.

define :method aoEnters(player :AoPlayer, house :AoHouse);
	;;; Make the player one of the occupants of the house
    [^player ^^(aoPlayers(house))] -> aoPlayers(house);

	;;; Tell the player that he/she has entered the house
	;;; Note that this uses a method of the same name, applied to
	;;; a different combination of types of objects
    aoEnters(player, aoStartloc(house));

	;;; Print out tracing information
    ['New player - named' %aoName(player)% in %aoName(aoRoom(player))%] =>
    aoIntroduce(player);
enddefine;

define :method aoLeaves(player :AoPlayer, house :AoHouse);
	;;; A method for a player to leave a house
	;;; Tell the player he/she is leaving
    aoLeaves(player, aoRoom(player));
	;;; Remove player from the occupants list of the house
    delete(player, aoPlayers(house))-> aoPlayers(house);
enddefine;

define :method aoEnters(player :AoPlayer, room :AoRoom);
	;;; Informing a player that he/she has entered a room
	;;; Note that this is invoked by another method of the same name:
	;;;    aoEnters(player :AoPlayer, house :AoHouse);

	;;; Inform the room of the new player in it, by adding to a list
    [^player ^^(aoPlayers(room))] -> aoPlayers(room);

	;;; Inform the player of the room currently occupied
    room -> aoRoom(player);
enddefine;

define :method aoLeaves(player :AoPlayer, room :AoRoom);
	;;; Remove the player from the list of occupants
    delete(player,aoPlayers(room)) -> aoPlayers(room);

	;;; And the room the player knows about temporarily becomes undefined
	undef -> aoRoom(player);
enddefine;

define :method aoTaken(physObj :AoPhysicalObject, room :AoRoom);
	;;; An object gets taken, so remove it from the list of objects
	;;; in its room
    delete(physObj, aoObjects(room)) -> aoObjects(room);
enddefine;


/* -------------- METHODS FOR PLAYERS -------------- */

;;; The default aoStrategy for a player is simply to do nothing.
;;; More specific strategies will be defined below for particular types of
;;; players

define :method aoStrategy(player :AoPlayer);
    [%aoName(player)% is making a plan] =>
enddefine;

;;; A default method for trying to move to another room
define :method aoTrymove(player :AoPlayer);
    lvars next_room;
    lvars procedure dir;

	;;; Randomly select a slot-access method corresponding to
	;;; one of the four directions out of a room, and assign it to dir
    oneof([%aoNorth, aoSouth, aoEast, aoWest%]) -> dir;

	;;; Now get the room that is in that direction (if there is one)
    dir(aoRoom(player)) -> next_room;

    if next_room then
		;;; print out who is moving from where to where
        [% aoName(player), 'moves from', aoName(aoRoom(player)),
			'to', aoName(next_room) %] =>
		;;; Make the player leave one room and enter another
        aoLeaves(player, aoRoom(player));
        aoEnters(player, next_room);
    else
        [%'Cannot move', aoName(player), 'in direction',
            pdprops(dir), 'from', aoName(aoRoom(player)) %] =>
    endif;
enddefine;

;;; More methods to do with agents

;;; A default Move strategy, which requires a minimum health level
define :method aoMove(player :AoPlayer);
    if aoHealth(player) > 10 then
        aoStrategy(player);
    endif;
enddefine;

;;; But what happens when the player dies
define :method aoDies(player :AoPlayer);
    [%aoName(player)% dies] =>

	;;; The player cannot be in the house any more
    aoLeaves(player, aoHouse(aoRoom(player)));
	;;; Should this also get the player to leave the room?
	;;; Should it change a slot in the player to indicate that it is
	;;; already dead, and therefore cannot be treated normally ?
enddefine;

define :method aoHit(agressor :AoPlayer, victim :AoPlayer);
	;;; Some nasty action to be reported
    [%aoName(agressor)% hits %aoName(victim)%] =>

	;;; And its consequences stored -- reduced health
    aoHealth(victim)/2.0 -> aoHealth(victim);

	;;; And possible death?
    if aoHealth(victim) <= 10 then
        aoDies(victim);
    endif;
enddefine;

/*
A possible exercise might be to have a class of players that are AoBrave.
Then hitting them might provoke hitting back (sometimes, or always).
*/


define :method aoTake(player :AoPlayer, object :AoPhysicalObject);

	;;; Print tracing information
    [% aoName(player), 'takes',  aoName(object) %] =>

	;;; Record the change
    aoTaken(object, aoRoom(player));
    [^object ^^( aoPossessions(player) )]-> aoPossessions(player);

	;;; Print more tracing information: number of possitions of player
    [% aoName(player), 'now owns',
         length(aoPossessions(player)), 'possessions' %] ==>
    ;;; You could alter that to print out the possession names also.

enddefine;


;;; METHODS FOR INTRODUCING NEW PLAYERS

;;; First a utility printing procedure
define print_introduction(playername, string);
	lvars playername, string;
	['Player' ^playername ^string] =>
enddefine;

;;; First the default method for normal players

define :method aoIntroduce(player :AoPlayer);
	;;; Default introduce method for players
    print_introduction(aoName(player), 'is normal.')
enddefine;

;;; Methods for introducing other types of player

define :method aoIntroduce(player :AoGreedy);
    print_introduction(aoName(player), 'is greedy.')
enddefine;

define :method aoIntroduce(player :AoVicious);
    print_introduction(aoName(player), 'is vicious.')
enddefine;


;;; SPECIFIC STRATEGY METHODS FOR SPECIAL TYPES OF PLAYERS

;;; Different types of players have different strategies

define :method aoStrategy(greedy :AoGreedy);
    ;;; If there's an object in the room, take it
    lvars objects_in_room = aoObjects(aoRoom(greedy));

    vars an_object; ;;; a pattern variable, used with the matcher

    if objects_in_room matches [?an_object  == ] then
        aoTake(greedy, an_object);
    else
		;;; Nothing to take, so try to move somewhere else
        aoTrymove(greedy);
    endif;
enddefine;

define :method aoStrategy( vicious :AoVicious);
    /*Try to find a player to hit, or move */

    lvars potential_victims, victim;

	;;; Make a list of everyone else in the room
	;;; First get the complete list
    aoPlayers(aoRoom(vicious)) -> potential_victims;
	;;; then delete this player
    delete(vicious, potential_victims) -> potential_victims;
	;;; should dead players also be deleted ???

    unless potential_victims == [] then
		;;; select a victim and try hitting
        oneof(potential_victims) -> victim;
        aoHit(vicious, victim);
    else
		;;; no potential victims here, so go looking elsewhere
        aoTrymove(vicious);
    endunless;
enddefine;


;;; TOP LEVEL PROCEDURE FOR THE SIMULATION

define :method aoPlay(house :AoHouse);
    lvars player;

    until aoPlayers(house) ==  [] do
        for player in aoPlayers(house)do
            aoMove(player);
        endfor;
    enduntil;
    [No more players] =>
enddefine;


/*

Now set up a house containing rooms with the following spatial relationships:

                     pit
                      ^
                      |
   swamp <------> slimy_dungeon

*/

define :method aoSetup(house :AoHouse);
    /*This sets up a house with rooms, objects, and players. */

    lvars room1, room2, startloc, player, players;

    "test_house" -> aoName(house);

    ;;;SET UP THREE ROOMS
    ;;;create a room, make it the house's aoStartloc
    instance AoRoom;
        aoName = "slimy_dungeon";
        aoObjects =
        [%instance AoPhysicalObject; aoName = "broken_bucket"; endinstance %];
        aoHouse = house;
    endinstance
        ->> aoStartloc(house) -> startloc; ;;; use startloc again below

    instance AoRoom;
        aoName = "swamp";
        aoObjects =
			[%
			 instance AoPhysicalObject;
				aoName = "shiny_jewels";
        		aoValue = 500;
        	 endinstance%];
        aoEast = startloc;
        aoHouse = house;
    endinstance -> room1;

    instance AoRoom;
        aoName = "pit";
        aoSouth = startloc;
        aoHouse = house;
    endinstance -> room2;

    ;;; set some more of startloc's fields
    ;;; Notice that this affects the house's startloc
    room1 -> aoWest(startloc);
    room2 -> aoNorth(startloc);

    [^startloc ^room1 ^room2]-> aoRooms(house);

    ;;;NOW MAKE THREE PLAYERS
    [%
        instance AoPlayer;
            aoName = "lazy";
        endinstance;

        instance AoGreedy;
            aoName = "seeker";
        endinstance;

        instance AoVicious;
            aoName = "nasty";
        	endinstance
    %] -> players;

    for player in players do
		;;; Make each player enter the house. This will cause an
		;;; introduction to be printed out
        aoEnters(player, house);
    endfor;
enddefine;


/* --------------  METHODS FOR PRINTING  -------------- */

;;; We can redefine objectclass print_instance method for objects
;;;		used in this adventure game.

;;; print_instance is the default class_print of all classes

define :method print_instance(player :AoPlayer);
    pr('<AoPlayer aoName:');
    pr(aoName(player));
    pr(' aoRoom:');
    pr(aoName(aoRoom(player)));
    pr('>')
enddefine;

define :method print_instance(room :AoRoom);
    lvars player;
    pr('<AoRoom aoName:');
    for player in aoPlayers(room)do
        print_instance(player);
    endfor;
	;;; A possible extension would be to enhance this to
    ;;; print the adjoining rooms as well as the players
    pr('>')
enddefine;

/*

Testing the program

Load the following line to run the simulation:

vars ao_house = newAoHouse();
aoSetup(ao_house);
aoPlay(ao_house);

You will need to interrupt aoPlay or it will go on forever.

INDEX OF OBJECTCLASS DEFINITIONS, METHODS AND PROCEDURES
    (Select desired one then type "ENTER g define")

CONTENTS (define)

 define :class AoObject;
 define :class AoHouse;
 define :class AoPhysicalObject;
 define :class AoRoom;
 define :class AoPlayer;
 define :class AoGreedy;
 define :class AoVicious;
 define :method aoEnters(player :AoPlayer, house :AoHouse);
 define :method aoLeaves(player :AoPlayer, house :AoHouse);
 define :method aoEnters(player :AoPlayer, room :AoRoom);
 define :method aoLeaves(player :AoPlayer, room :AoRoom);
 define :method aoTaken(physObj :AoPhysicalObject, room :AoRoom);
 define :method aoStrategy(player :AoPlayer);
 define :method aoTrymove(player :AoPlayer);
 define :method aoMove(player :AoPlayer);
 define :method aoDies(player :AoPlayer);
 define :method aoHit(agressor :AoPlayer, victim :AoPlayer);
 define :method aoTake(player :AoPlayer, object :AoPhysicalObject);
 define print_introduction(playername, string);
 define :method aoIntroduce(player :AoPlayer);
 define :method aoIntroduce(player :AoGreedy);
 define :method aoIntroduce(player :AoVicious);
 define :method aoStrategy(greedy :AoGreedy);
 define :method aoStrategy( vicious :AoVicious);
 define :method aoPlay(house :AoHouse);
 define :method aoSetup(house :AoHouse);
 define :method print_instance(player :AoPlayer);
 define :method print_instance(room :AoRoom);

*/

/* --- Revision History ---------------------------------------------------
--- Aaron Sloman, Dec  9 1995
	Added some more comments about the dead player bug.
 */
